@@ -42,6 +42,16 @@ enum {
 	FIO_FSERVICE_SEQ	= 3,
 };
 
+/*
+ * No pre-allocation when laying down files, or call posix_fallocate(), or
+ * call fallocate() with FALLOC_FL_KEEP_SIZE set.
+ */
+enum fio_fallocate_mode {
+	FIO_FALLOCATE_NONE	= 1,
+	FIO_FALLOCATE_POSIX	= 2,
+	FIO_FALLOCATE_KEEP_SIZE	= 3,
+};
+
 /*
  * Each thread_data structure has a number of files associated with it,
  * this structure holds state information for a single file.
@@ -97,9 +107,9 @@ struct fio_file {
 	 * block map for random io
 	 */
 	unsigned long *file_map;
-	unsigned int num_maps;
-	unsigned int last_free_lookup;
-	unsigned int failed_rands;
+	unsigned long num_maps;
+	unsigned long last_free_lookup;
+	unsigned failed_rands;
 
 	int references;
 	enum fio_file_flags flags;
